from datetime import timedelta
from types import MethodType

import pandas as pd
import pytest

from vivarium.framework.results import ResultsInterface, ResultsManager

from .mocks import BASE_POPULATION
from .mocks import CATEGORIES as HOUSES
from .mocks import FAMILIARS, mock_get_value


def _silly_aggregator(_: pd.DataFrame) -> float:
    return 1.0


@pytest.mark.parametrize(
    (
        "name, pop_filter, aggregator_columns, aggregator, requires_columns, requires_values,"
        " additional_stratifications, excluded_stratifications, when"
    ),
    [
        (
            "living_person_time",
            'alive == "alive" and undead == False',
            [],
            _silly_aggregator,
            [],
            [],
            [],
            [],
            "collect_metrics",
        ),
        (
            "undead_person_time",
            "undead == True",
            [],
            _silly_aggregator,
            [],
            [],
            [],
            [],
            "time_step__prepare",
        ),
        (
            "undead_person_time",
            "undead == True",
            [],
            _silly_aggregator,
            [],
            ["fake_pipeline", "another_fake_pipeline"],
            [],
            [],
            "time_step__prepare",
        ),
    ],
    ids=["valid_on_collect_metrics", "valid_on_time_step__prepare", "valid_pipelines"],
)
def test_register_observation(
    mocker,
    name,
    pop_filter,
    aggregator_columns,
    aggregator,
    requires_columns,
    requires_values,
    additional_stratifications,
    excluded_stratifications,
    when,
):
    mgr = ResultsManager()
    interface = ResultsInterface(mgr)
    builder = mocker.Mock()
    builder.configuration.stratification.default = []
    # Set up mock builder with mocked get_value call for Pipelines
    mocker.patch.object(builder, "value.get_value")
    builder.value.get_value = MethodType(mock_get_value, builder)
    mgr.setup(builder)
    assert len(interface._manager._results_context.observations) == 0
    interface.register_observation(
        name,
        pop_filter,
        aggregator_columns,
        aggregator,
        requires_columns,
        requires_values,
        additional_stratifications,
        excluded_stratifications,
    )
    assert len(interface._manager._results_context.observations) == 1


def test_register_observations(mocker):
    mgr = ResultsManager()
    interface = ResultsInterface(mgr)
    builder = mocker.Mock()
    builder.configuration.stratification.default = []
    mgr.setup(builder)

    assert len(interface._manager._results_context.observations) == 0
    interface.register_observation(
        "living_person_time",
        'alive == "alive" and undead == False',
        [],
        _silly_aggregator,
        [],
        [],
        [],
        [],
        "collect_metrics",
    )
    assert len(interface._manager._results_context.observations) == 1
    interface.register_observation(
        "undead_person_time",
        "undead == True",
        [],
        _silly_aggregator,
        [],
        [],
        [],
        [],
        "time_step__prepare",
    )
    assert len(interface._manager._results_context.observations) == 2


def test_unhashable_pipeline(mocker):
    mgr = ResultsManager()
    interface = ResultsInterface(mgr)
    builder = mocker.Mock()
    builder.configuration.stratification.default = []
    mgr.setup(builder)

    assert len(interface._manager._results_context.observations) == 0
    with pytest.raises(TypeError, match="unhashable"):
        interface.register_observation(
            "living_person_time",
            'alive == "alive" and undead == False',
            [],
            _silly_aggregator,
            [],
            [["bad", "unhashable", "thing"]],  # unhashable first element
            [],
            [],
            "collect_metrics",
        )


def mock__prepare_population(self, event):
    """Return a mock population in the vein of ResultsManager._prepare_population"""
    # Generate population DataFrame
    population = BASE_POPULATION.copy()

    # Mock out some extra columns that would be produced by the manager's _prepare_population() method
    population["current_time"] = pd.Timestamp(year=2045, month=1, day=1, hour=12)
    population["step_size"] = timedelta(days=28)
    population["event_time"] = pd.Timestamp(year=2045, month=1, day=1, hour=12) + timedelta(
        days=28
    )
    return population


def test_integration_full_observation(mocker):
    """Test the full interface lifecycle of adding an observation and simulate a `collect_metrics` event."""
    # Create interface
    mgr = ResultsManager()
    results_interface = ResultsInterface(mgr)
    builder = mocker.Mock()
    builder.configuration.stratification.default = []
    mgr.setup(builder)

    # register stratifications
    results_interface.register_stratification("house", HOUSES, None, True, ["house"], [])
    results_interface.register_stratification(
        "familiar", FAMILIARS, None, True, ["familiar"], []
    )

    mock_aggregator = mocker.Mock()
    another_mock_aggregator = mocker.Mock()

    results_interface.register_observation(
        "a_measure",
        "tracked==True",
        None,
        mock_aggregator,
        ["house", "familiar"],
        [],
        ["house", "familiar"],
        [],
        "collect_metrics",
    )
    # register observation
    results_interface.register_observation(
        "another_measure",
        "tracked==True",
        None,
        another_mock_aggregator,
        ["house", "familiar"],
        [],
        ["house", "familiar"],
        [],
        "time_step__prepare",
    )

    # Mock in mgr._prepare_population to return population table, event
    mocker.patch.object(mgr, "_prepare_population")
    mgr._prepare_population = MethodType(mock__prepare_population, mgr)

    mock_aggregator.assert_not_called()
    another_mock_aggregator.assert_not_called()

    # Fake a timestep
    mock_event = mocker.Mock()
    mgr.gather_results("collect_metrics", mock_event)

    mock_aggregator.assert_called()  # Observation aggregator that should have been called
    another_mock_aggregator.assert_not_called()  # Observation aggregator that should not



import pandas as pd
from scipy import spatial


class Neighbors:
    configuration_defaults = {"neighbors": {"radius": 10}}

    def __init__(self):
        self.name = "Neighbors"

    def setup(self, builder):
        self.radius = builder.configuration.neighbors.radius

        self.neighbors_calculated = False
        self._neighbors = pd.Series()
        self.neighbors = builder.value.register_value_producer(
            "neighbors", source=self.get_neighbors
        )

        builder.population.initializes_simulants(self.on_create_simulants)
        self.population_view = builder.population.get_view(["x", "y"])

        builder.event.register_listener("time_step", self.on_time_step)

    def on_create_simulants(self, pop_data):
        self._neighbors = pd.Series([[]] * len(pop_data.index), index=pop_data.index)

    def on_time_step(self, event):
        self.neighbors_calculated = False

    def get_neighbors(self, index):
        if not self.neighbors_calculated:
            self.calculate_neighbors()
        return self._neighbors[index]

    def calculate_neighbors(self):
        # Reset our list of neighbors
        pop = self.population_view.get(self._neighbors.index)
        self._neighbors = pd.Series([[]] * len(pop), index=pop.index)

        tree = spatial.KDTree(pop)

        # Iterate over each pair of simulates that are close together.
        for boid_1, boid_2 in tree.query_pairs(self.radius):
            # .iloc is used because query_pairs uses 0,1,... indexing instead of pandas.index
            self._neighbors.iloc[boid_1].append(self._neighbors.index[boid_2])
            self._neighbors.iloc[boid_2].append(self._neighbors.index[boid_1])



import numpy as np
import pandas as pd


class Location:
    configuration_defaults = {
        "location": {
            "width": 1000,  # Width of our field
            "height": 1000,  # Height of our field
        }
    }

    def __init__(self):
        self.name = "location"

    def setup(self, builder):
        self.width = builder.configuration.location.width
        self.height = builder.configuration.location.height

        columns_created = ["x", "vx", "y", "vy"]
        builder.population.initializes_simulants(self.on_create_simulants, columns_created)
        self.population_view = builder.population.get_view(columns_created)

    def on_create_simulants(self, pop_data):
        count = len(pop_data.index)
        # Start clustered in the center with small random velocities
        new_population = pd.DataFrame(
            {
                "x": self.width * (0.4 + 0.2 * np.random.random(count)),
                "y": self.height * (0.4 + 0.2 * np.random.random(count)),
                "vx": -0.5 + np.random.random(count),
                "vy": -0.5 + np.random.random(count),
            },
            index=pop_data.index,
        )
        self.population_view.update(new_population)



import warnings
from collections import defaultdict
from typing import Callable, Dict, List, Tuple

import pandas as pd

from vivarium.framework.results.exceptions import ResultsConfigurationError
from vivarium.framework.results.stratification import Stratification


class ResultsContext:
    """
    Manager context for organizing observations and the stratifications they require.

    This context object is wholly contained by the manager :class:`vivarium.framework.results.manager.ResultsManger`.
    Stratifications can be added to the context through the manager via the
    :meth:`vivarium.framework.results.context.ResultsContext.add_observation` method.
    """

    def __init__(self):
        self.default_stratifications: List[str] = []
        self.stratifications: List[Stratification] = []
        # keys are event names: [
        #     "time_step__prepare",
        #     "time_step",
        #     "time_step__cleanup",
        #     "collect_metrics",
        # ]
        # values are dicts with
        #     key (filter, grouper)
        #     value (measure, aggregator_sources, aggregator, additional_keys)
        self.observations = defaultdict(lambda: defaultdict(list))

    @property
    def name(self):
        return "results_context"

    def setup(self, builder):
        self.logger = builder.logging.get_logger(self.name)

    # noinspection PyAttributeOutsideInit
    def set_default_stratifications(self, default_grouping_columns: List[str]):
        if self.default_stratifications:
            raise ResultsConfigurationError(
                "Multiple calls are being made to set default grouping columns "
                "for results production."
            )
        self.default_stratifications = default_grouping_columns

    def add_stratification(
        self,
        name: str,
        sources: List[str],
        categories: List[str],
        mapper: Callable,
        is_vectorized: bool,
    ):
        """Add a stratification to the context.

        Parameters
        ----------
        name
            Name of the of the column created by the stratification.
        sources
            A list of the columns and values needed for the mapper to determinate
            categorization.
        categories
            List of string values that the mapper is allowed to output.
        mapper
            A callable that emits values in `categories` given inputs from columns
            and values in the `requires_columns` and `requires_values`, respectively.
        is_vectorized
            `True` if the mapper function expects a `DataFrame`, and `False` if it
            expects a row of the `DataFrame` and should be used by calling :func:`df.apply`.


        Returns
        ------
        None

        """
        if len([s.name for s in self.stratifications if s.name == name]):
            raise ValueError(f"Name `{name}` is already used")
        stratification = Stratification(name, sources, categories, mapper, is_vectorized)
        self.stratifications.append(stratification)

    def add_observation(
        self,
        name: str,
        pop_filter: str,
        aggregator_sources: List[str],
        aggregator: Callable[[pd.DataFrame], float],
        additional_stratifications: List[str] = (),
        excluded_stratifications: List[str] = (),
        when: str = "collect_metrics",
        **additional_keys: str,
    ):
        stratifications = self._get_stratifications(
            additional_stratifications, excluded_stratifications
        )
        self.observations[when][(pop_filter, stratifications)].append(
            (name, aggregator_sources, aggregator, additional_keys)
        )

    def gather_results(self, population: pd.DataFrame, event_name: str) -> Dict[str, float]:
        # Optimization: We store all the producers by pop_filter and stratifications
        # so that we only have to apply them once each time we compute results.
        for stratification in self.stratifications:
            population = stratification(population)

        for (pop_filter, stratifications), observations in self.observations[
            event_name
        ].items():
            # Results production can be simplified to
            # filter -> groupby -> aggregate in all situations we've seen.
            if pop_filter:
                filtered_pop = population.query(pop_filter)
            else:
                filtered_pop = population
            if filtered_pop.empty:
                yield {}
            else:
                if not len(list(stratifications)):  # Handle situation of no stratifications
                    pop_groups = filtered_pop.groupby(lambda _: True)
                else:
                    pop_groups = filtered_pop.groupby(list(stratifications))

                for measure, aggregator_sources, aggregator, additional_keys in observations:
                    if aggregator_sources:
                        aggregates = (
                            pop_groups[aggregator_sources].apply(aggregator).fillna(0.0)
                        )
                    else:
                        aggregates = pop_groups.apply(aggregator)

                    # Ensure we are dealing with a single column of formattable results
                    if isinstance(aggregates, pd.DataFrame):
                        aggregates = aggregates.squeeze(axis=1)
                    if not isinstance(aggregates, pd.Series):
                        raise TypeError(
                            f"The aggregator return value is a {type(aggregates)} and could not be "
                            "made into a pandas.Series. This is probably not correct."
                        )

                    # Keep formatting all in one place.
                    yield self._format_results(
                        measure,
                        aggregates,
                        bool(len(list(stratifications))),
                        **additional_keys,
                    )

    def _get_stratifications(
        self,
        additional_stratifications: List[str] = (),
        excluded_stratifications: List[str] = (),
    ) -> Tuple[str, ...]:
        stratifications = list(
            set(self.default_stratifications) - set(excluded_stratifications)
            | set(additional_stratifications)
        )
        # Makes sure measure identifiers have fields in the same relative order.
        return tuple(sorted(stratifications))

    @staticmethod
    def _format_results(
        measure: str,
        aggregates: pd.Series,
        has_stratifications: bool,
        **additional_keys: str,
    ) -> Dict[str, float]:
        results = {}
        # Simpler formatting if we don't have stratifications
        if not has_stratifications:
            return {measure: aggregates.squeeze()}

        # First we expand the categorical index over unobserved pairs.
        # This ensures that the produced results are always the same length.
        if isinstance(aggregates.index, pd.MultiIndex):
            idx = pd.MultiIndex.from_product(
                aggregates.index.levels, names=aggregates.index.names
            )
        else:
            idx = aggregates.index
        data = pd.Series(data=0, index=idx)
        data.loc[aggregates.index] = aggregates

        def _format(field, param):
            """Format of the measure identifier tokens into FIELD_param."""
            return f"{str(field).upper()}_{param}"

        for categories, val in data.items():
            if isinstance(categories, str):  # handle single stratification case
                categories = [categories]
            key = "_".join(
                [_format("measure", measure)]
                + [
                    _format(field, category)
                    for field, category in zip(data.index.names, categories)
                ]
                # Sorts additional_keys by the field name.
                + [
                    _format(field, category)
                    for field, category in sorted(additional_keys.items())
                ]
            )
            results[key] = val
        return results



import numpy as np
import pandas as pd


class Population:
    configuration_defaults = {
        "population": {
            "colors": ["red", "blue"],
        }
    }

    def __init__(self):
        self.name = "population"

    def setup(self, builder):
        self.colors = builder.configuration.population.colors

        columns_created = ["color", "entrance_time"]
        builder.population.initializes_simulants(
            self.on_initialize_simulants, columns_created
        )
        self.population_view = builder.population.get_view(columns_created)

    def on_initialize_simulants(self, pop_data):
        new_population = pd.DataFrame(
            {
                "color": np.random.choice(self.colors, len(pop_data.index)),
                "entrance_time": pop_data.creation_time,
            },
            index=pop_data.index,
        )
        self.population_view.update(new_population)
